package vexriscv.demo

import spinal.core._
import spinal.lib._
import spinal.lib.bus.amba3.apb.{Apb3, Apb3Config, Apb3Gpio, Apb3SlaveFactory}
import spinal.lib.memory.sdram.SdramGeneration._
import spinal.lib.memory.sdram.SdramLayout
import spinal.lib.bus.amba4.axi._
import spinal.lib.com.uart.UartCtrlMemoryMappedConfig
import spinal.lib.io.TriStateArray
import vexriscv.VexRiscv
import vexriscv.plugin.Plugin

import scala.collection.mutable.ArrayBuffer

object MT41K256M16TW {
  def layout = SdramLayout(
    generation = DDR3,
    bankWidth = 3,
    columnWidth = 10,
    rowWidth = 15,
    dataWidth = 16
  )
}

//add interrupt controller function
case class Apb3GpioIntCtrl(gpioWidth: Int, withReadSync : Boolean, debunceTimeStep : TimeNumber = 20 ms) extends Component {
  val io = new Bundle {
    val apb  = slave(Apb3(Apb3Config(addressWidth = 5, dataWidth = 32)))
    val gpio = master(TriStateArray(gpioWidth bits))
    val value = out Bits(gpioWidth bits)
    val intSignal = out Bits(gpioWidth bits)
  }

  io.value := (if(withReadSync) BufferCC(io.gpio.read) else io.gpio.read)

  val intEnable = Bits(gpioWidth bits)
  val intPolarity = Bits(gpioWidth bits) // 0:high level, 1:low level

  val btnDebunce = BtnDebunce(gpioWidth, debunceTimeStep)
  btnDebunce.io.btnIn := (io.gpio.read ^ intPolarity) & (~io.gpio.writeEnable) //only accept interrupt signal when at input status
  val debunceOut = Reg(Bits(gpioWidth bits)) init(0) //used to generate level signal for interrupt
  io.intSignal := debunceOut & intEnable
  //debunceOut := btnDebunce.io.debunceOut.asBools.map(_.rise()).asBits()

  val intClear = Bits(gpioWidth bits)
  intClear := U(0, gpioWidth bits).asBits
  var conditionSet = btnDebunce.io.debunceOut.asBools.map(_.rise())
  var index = 0
  for(index <- 0 until gpioWidth){
    when(intClear(index) || ~intEnable(index)){
      debunceOut(index) := False
    }.elsewhen(conditionSet(index)) {
      debunceOut(index) := True
    }
  }

  val ctrl = Apb3SlaveFactory(io.apb)
  ctrl.read(io.value, 0)
  ctrl.driveAndRead(io.gpio.write, 4)
  ctrl.driveAndRead(io.gpio.writeEnable, 8)
  ctrl.setOnSet(intClear, 12) //used to clean the interrupt signal
  ctrl.driveAndRead(intEnable, 16)
  ctrl.driveAndRead(intPolarity, 20)

  //initial value setting
  io.gpio.writeEnable.getDrivingReg() init(0)
  intPolarity.getDrivingReg() init(0)
  intEnable.getDrivingReg() init(0)
}

// Used to eliminate button bunce
case class BtnDebunce(dataWidth : Int, timeStep : TimeNumber = 20 ms) extends Component {
  val io = new Bundle{
    val btnIn = in Bits(dataWidth bits)
    val debunceOut = out Bits(dataWidth bits)
  }

  val counter = Timeout(timeStep)
  val regBtnIn = Reg(UInt(dataWidth bits)) init(0)

  when(counter.counter.willOverflow){
    regBtnIn := io.btnIn.asUInt
  }

  io.debunceOut := regBtnIn.asBits

}

case class TcmConfig(tcmUsed : Boolean = false,
                      ramAsBlackbox : Boolean = false,
                      hexInit: String = null,
                      tcmRamAddress: BigInt = 0x20000000,
                      tcmRamSize : BigInt = 0x1000)

case class Ddr3CtrlIF(sdramLayout: SdramLayout) extends Bundle {
  //ddr3 interface
  val ddr3_addr = out Bits (sdramLayout.rowWidth bits) //val ddr3_addr = out Bits(15 bits)
  val ddr3_ba = out Bits (sdramLayout.bankWidth bits) //val ddr3_ba = out Bits(3 bits)
  val ddr3_cas_n = out Bool()
  val ddr3_ck_n = out Bool()
  val ddr3_ck_p = out Bool()
  val ddr3_cke = out Bool()
  val ddr3_ras_n = out Bool()
  val ddr3_reset_n = out Bool()
  val ddr3_we_n = out Bool()
  val ddr3_dq = inout(Analog(Bits (sdramLayout.dataWidth bits)))  //val ddr3_dq = inout Bits(16 bits)
  val ddr3_dqs_n = inout(Analog(Bits ((sdramLayout.dataWidth / 8) bits)))
  val ddr3_dqs_p = inout(Analog(Bits ((sdramLayout.dataWidth / 8) bits)))
  val ddr3_dm = out Bits ((sdramLayout.dataWidth / 8) bits)
  val ddr3_odt = out Bool()
}

class SystemClkGen() extends BlackBox{
  val io = new Bundle{
    val clk_out1 = out Bits(1 bits)
    val resetn = in Bits(1 bits)
    val locked = out Bits(1 bits)
    val clk_in1 = in Bits(1 bits)
  }
  noIoPrefix()
}

class Ddr3Ctrl(sdramLayout: SdramLayout,axi4Config: Axi4Config) extends BlackBox{
  val io = new Bundle{
    //ddr3 interface
    val ddr3IF = Ddr3CtrlIF(sdramLayout)
//    val ddr3_addr = out Bits(sdramLayout.rowWidth bits) //val ddr3_addr = out Bits(15 bits)
//    val ddr3_ba = out Bits(sdramLayout.bankWidth bits)  //val ddr3_ba = out Bits(3 bits)
//    val ddr3_cas_n = out Bool()
//    val ddr3_ck_n = out Bool()
//    val ddr3_ck_p = out Bool()
//    val ddr3_cke = out Bool()
//    val ddr3_ras_n = out Bool()
//    val ddr3_reset_n = out Bool()
//    val ddr3_we_n = out Bool()
//    val ddr3_dq = inout Bits(sdramLayout.dataWidth bits) //val ddr3_dq = inout Bits(16 bits)
//    val ddr3_dqs_n = inout Bits((sdramLayout.dataWidth/2) bits)
//    val ddr3_dqs_p = inout Bits((sdramLayout.dataWidth/2) bits)
//    val init_calib_complete = out Bool()
//    val ddr3_dm = out Bits((sdramLayout.dataWidth/2) bits)
//    val ddr3_odt = out Bool()

    //vivado interface
    val init_calib_complete = out Bool() //used to reset cpu
    val ui_clk = out Bool() //bus and cpu clock
    val ui_clk_sync_rst = out Bool()  //used to reset bus
    val mmcm_locked = out Bool()  //not used
    val aresetn = in Bool() //reset axi bus
    val app_sr_req = in Bool()
    val app_ref_req = in Bool()
    val app_zq_req = in Bool()
    val app_sr_active = out Bool()
    val app_ref_ack = out Bool()
    val app_zq_ack = out Bool()
    val sys_clk_i = in Bool()
    val sys_rst = in Bool()

    //axi4 interface
    val s_axi_awid = in Bits(1 bits)
    val s_axi_awaddr = in Bits(axi4Config.addressWidth bits)
    val s_axi_awlen = in Bits(8 bits)
    val s_axi_awsize = in Bits(3 bits)
    val s_axi_awburst = in Bits(2 bits)
    val s_axi_awlock = in Bits(1 bits)
    val s_axi_awcache = in Bits(4 bits)
    val s_axi_awprot = in Bits(3 bits)
    val s_axi_awqos = in Bits(4 bits)
    val s_axi_awvalid = in Bits(1 bits)
    val s_axi_awready = out Bits(1 bits)
    val s_axi_wdata = in Bits(axi4Config.dataWidth bits)
    val s_axi_wstrb = in Bits(4 bits)
    val s_axi_wlast = in Bits(1 bits)
    val s_axi_wvalid = in Bits(1 bits)
    val s_axi_wready = out Bits(1 bits)
    val s_axi_bid = out Bits(1 bits)
    val s_axi_bresp = out Bits(2 bits)
    val s_axi_bvalid = out Bits(1 bits)
    val s_axi_bready = in Bits(1 bits)
    val s_axi_arid = in Bits(1 bits)
    val s_axi_araddr = in Bits(axi4Config.addressWidth bits)
    val s_axi_arlen = in Bits(8 bits)
    val s_axi_arsize = in Bits(3 bits)
    val s_axi_arburst = in Bits(2 bits)
    val s_axi_arlock = in Bits(1 bits)
    val s_axi_arcache = in Bits(4 bits)
    val s_axi_arprot = in Bits(3 bits)
    val s_axi_arqos = in Bits(4 bits)
    val s_axi_arvalid = in Bits(1 bits)
    val s_axi_arready = out Bits(1 bits)
    val s_axi_rid = out Bits(1 bits)
    val s_axi_rdata = out Bits(axi4Config.dataWidth bits)
    val s_axi_rresp = out Bits(2 bits)
    val s_axi_rlast = out Bits(1 bits)
    val s_axi_rvalid = out Bits(1 bits)
    val s_axi_rready = in Bits(1 bits)
  }
  noIoPrefix()

  //⽤函数给⿊盒中所有信号重命名
  private def renameIO(): Unit = {
    io.flatten.foreach(bt => {
      if (bt.getName().contains("ddr3IF")) bt.setName(bt.getName().replace("ddr3IF_", ""))
    })
  }
    //在创建模块后执⾏重命名函数
  addPrePopTask(() => renameIO())
}

class VivadoMigWrapper(sdramLayout: SdramLayout) extends Component{
  val ddr3Axi4config = Axi4Config(addressWidth=(sdramLayout.rowWidth+sdramLayout.columnWidth+sdramLayout.bankWidth+sdramLayout.dataWidth/8-1),
    dataWidth=32, useId=false).toFullConfig()
  val axi4config = Axi4Config(addressWidth = 32, dataWidth = 32, useId = false).toFullConfig()

  val io = new Bundle {
    //Clocks / reset
    val systemClk = in Bool()
    val systemResetn = in Bool()
    val busClk = out Bool()
    val busReset = out Bool()
    val cpuReset = out Bool()

    //DDR3 IF
    val ddr3IF = Ddr3CtrlIF(sdramLayout)

    //axi4 IF
    val s_axi4Bus = slave(Axi4(axi4config))
  }

  noIoPrefix()

  //⽤函数给⿊盒中所有信号重命名
  private def renameIO(): Unit = {
    io.flatten.foreach(bt => {
      if (bt.getName().contains("ddr3IF")) bt.setName(bt.getName().replace("ddr3IF_", ""))
    })
  }
  //在创建模块后执⾏重命名函数
  addPrePopTask(() => renameIO())

  // for ddr3 IF
  val ddr3Ctrl = new Ddr3Ctrl(sdramLayout , ddr3Axi4config)
  io.ddr3IF <> ddr3Ctrl.io.ddr3IF

  // for axi4 IF
  val addressWidth = (sdramLayout.rowWidth+sdramLayout.columnWidth+sdramLayout.bankWidth+sdramLayout.dataWidth/8-1)
  ddr3Ctrl.io.s_axi_awid    <> io.s_axi4Bus.aw.id.asBits
  ddr3Ctrl.io.s_axi_awaddr  <> io.s_axi4Bus.aw.addr.asBits(0 to addressWidth-1)
  ddr3Ctrl.io.s_axi_awlen   <> io.s_axi4Bus.aw.len.asBits
  ddr3Ctrl.io.s_axi_awsize  <> io.s_axi4Bus.aw.size.asBits
  ddr3Ctrl.io.s_axi_awburst <> io.s_axi4Bus.aw.burst
  ddr3Ctrl.io.s_axi_awlock  <> io.s_axi4Bus.aw.lock
  ddr3Ctrl.io.s_axi_awcache <> io.s_axi4Bus.aw.cache
  ddr3Ctrl.io.s_axi_awprot  := io.s_axi4Bus.aw.prot
  ddr3Ctrl.io.s_axi_awqos   := io.s_axi4Bus.aw.qos
  ddr3Ctrl.io.s_axi_awvalid := io.s_axi4Bus.aw.valid.asBits
  io.s_axi4Bus.aw.ready     := ddr3Ctrl.io.s_axi_awready.asBool
  ddr3Ctrl.io.s_axi_wdata   := io.s_axi4Bus.w.data
  ddr3Ctrl.io.s_axi_wstrb   := io.s_axi4Bus.w.strb
  ddr3Ctrl.io.s_axi_wlast   := io.s_axi4Bus.w.last.asBits
  ddr3Ctrl.io.s_axi_wvalid  := io.s_axi4Bus.w.valid.asBits
  io.s_axi4Bus.w.ready      := ddr3Ctrl.io.s_axi_wready.asBool
  io.s_axi4Bus.b.id         := ddr3Ctrl.io.s_axi_bid.asUInt
  io.s_axi4Bus.b.resp       := ddr3Ctrl.io.s_axi_bresp
  io.s_axi4Bus.b.valid      := ddr3Ctrl.io.s_axi_bvalid.asBool
  ddr3Ctrl.io.s_axi_bready  := io.s_axi4Bus.b.ready.asBits
  ddr3Ctrl.io.s_axi_arid    := io.s_axi4Bus.ar.id.asBits
  ddr3Ctrl.io.s_axi_araddr  := io.s_axi4Bus.ar.addr.asBits(0 to addressWidth-1)
  ddr3Ctrl.io.s_axi_arlen   := io.s_axi4Bus.ar.len.asBits
  ddr3Ctrl.io.s_axi_arsize  := io.s_axi4Bus.ar.size.asBits
  ddr3Ctrl.io.s_axi_arburst := io.s_axi4Bus.ar.burst
  ddr3Ctrl.io.s_axi_arlock  := io.s_axi4Bus.ar.lock
  ddr3Ctrl.io.s_axi_arcache := io.s_axi4Bus.ar.cache
  ddr3Ctrl.io.s_axi_arprot  := io.s_axi4Bus.ar.prot
  ddr3Ctrl.io.s_axi_arqos   := io.s_axi4Bus.ar.qos
  ddr3Ctrl.io.s_axi_arvalid := io.s_axi4Bus.ar.valid.asBits
  io.s_axi4Bus.ar.ready     := ddr3Ctrl.io.s_axi_arready.asBool
  io.s_axi4Bus.r.id         := ddr3Ctrl.io.s_axi_rid.asUInt
  io.s_axi4Bus.r.data       := ddr3Ctrl.io.s_axi_rdata
  io.s_axi4Bus.r.resp       := ddr3Ctrl.io.s_axi_rresp
  io.s_axi4Bus.r.last       := ddr3Ctrl.io.s_axi_rlast.asBool
  io.s_axi4Bus.r.valid      := ddr3Ctrl.io.s_axi_rvalid.asBool
  ddr3Ctrl.io.s_axi_rready  := io.s_axi4Bus.r.ready.asBits



  //clock domain
  val uiClockDomainConfig = ClockDomainConfig(resetKind = SYNC, resetActiveLevel = HIGH)
  val uiClockDomain = ClockDomain.internal("uiClockDomain")
  uiClockDomain.clock := ddr3Ctrl.io.ui_clk
  uiClockDomain.reset := ddr3Ctrl.io.ui_clk_sync_rst
  val uiClockArea = new ClockingArea(uiClockDomain){
    val ui_clk_sync_rstn = Reg(Bool()) init(False)
    ui_clk_sync_rstn := True
  }
  ddr3Ctrl.io.aresetn := uiClockArea.ui_clk_sync_rstn
  //ui IF
  //ddr3Ctrl.io.aresetn <> io.systemResetn
  ddr3Ctrl.io.app_sr_req := False
  ddr3Ctrl.io.app_ref_req := False
  ddr3Ctrl.io.app_zq_req := False

  ddr3Ctrl.io.sys_clk_i <> io.systemClk
  ddr3Ctrl.io.sys_rst <> io.systemResetn

  io.cpuReset <> (~ddr3Ctrl.io.init_calib_complete | ddr3Ctrl.io.ui_clk_sync_rst)
  io.busClk <> ddr3Ctrl.io.ui_clk
  io.busReset <> ddr3Ctrl.io.ui_clk_sync_rst
}


object VivadoMigWrapper{
  def main(args: Array[String]) {
    SpinalVerilog(new VivadoMigWrapper(MT41K256M16TW.layout)).printPruned()
  }
}